import os

from _pytest.config import Config


class TrConfig:
    def __init__(self, config: Config):
        self._config = config
        self.start_client = config.getoption('--tr-reporting')
        self.deselect_tests = config.getoption('--tr_deselect_tests', False)
        self.run_id = os.getenv('TR_RUN_ID') or config.getini('tr_run_id')
        self.api_url = os.getenv('TR_API_URL') or config.getini('tr_api_url')
        self.user_email = os.getenv('TR_USER_EMAIL') or config.getini('tr_user_email')
        self.user_password = os.getenv('TR_USER_PASSWORD') or config.getini('tr_user_password')
        self.project_id = os.getenv('TR_PROJECT_ID') or config.getini('tr_project_id')



from typing import List, Iterable

from testrail_api import TestRailAPI

from pytest_testrail_integrator.config import TrConfig


class TrService:
    _tests: list

    def __init__(self, config):
        self._config: TrConfig = config.tr_config
        self._api = TestRailAPI(self._config.api_url, self._config.user_email, self._config.user_password)

    def is_test_run_available(self) -> bool:
        return bool(self._config.run_id)

    @property
    def tests(self):
        """
        Gets Tests from run if RunId is defined.
        :raises ValueError: When runId in configs is not defined.
        :return: List of test dicts.
        """
        if not hasattr(self, '_tests'):
            if not self._config.run_id:
                raise ValueError('RunId is not available.')
            self._tests = self._api.tests.get_tests(int(self._config.run_id))
        return self._tests

    def get_cases(self) -> List[int]:
        """
        Gets list of case ids within entire test run if defined.
        :return: List of integers.
        """
        return [x['case_id'] for x in self.tests]

    def upload_results(self, results):
        if results:
            return self._api.results.add_results_for_cases(int(self._config.run_id), results)

    def create_test_run(self, case_ids: Iterable[int]) -> int:
        """
        Creates new test with test cases matching with passed case_ids param.
        :param case_ids: List of int, testrail case ids.
        :return:
        """
        r = self._api.runs.add_run(
            self._config.project_id,
            name='Automated Test run',
            include_all=False,
            case_ids=list(case_ids)
        )
        self._config.run_id = r['id']  # guessing the key
        return 0



import enum

TR_MARKER_NAME = 'case'
TR_PASSED_TESTS_FLUSH_SIZE = 10


class TestrailStatus(enum.Enum):
    PASSED = 1
    BLOCKED = 2
    UNTESTED = 3
    RETEST = 4
    FAILED = 5
    IN_PROGRESS = 6
    SKIP = 7
    REOPENED = 8


class PytestStatus(enum.Enum):
    PASSED = 'passed'
    FAILED = 'failed'
    SKIPPED = 'skipped'


TESTRAIL_STATUS_PRIORITY = {
    TestrailStatus.PASSED: 1,
    TestrailStatus.SKIP: 2,
    TestrailStatus.FAILED: 3
}
PYTEST_TO_TESTRAIL_STATUS = {
    PytestStatus.PASSED: TestrailStatus.PASSED,
    PytestStatus.FAILED: TestrailStatus.FAILED,
    PytestStatus.SKIPPED: TestrailStatus.SKIP
}



import dataclasses
from typing import List, Union

from _pytest._code import ExceptionInfo
from _pytest._code.code import TerminalRepr
from _pytest.mark import Mark

from pytest_testrail_integrator.constants import PytestStatus


@dataclasses.dataclass
class ReportDTO:
    name: str
    nodeid: str
    markers: List[Mark]
    status: PytestStatus
    duration: float
    longrepr: Union[None, ExceptionInfo[BaseException], tuple[str, int, str], str, TerminalRepr]
    case_id: int



from _pytest.config.argparsing import Parser

from pytest_testrail_integrator.client import TrClient
from pytest_testrail_integrator.service import TrService
from pytest_testrail_integrator.config import TrConfig


def pytest_addoption(parser: Parser):
    group = parser.getgroup('testrail')
    group.addoption(
        '--tr-reporting',
        action='store_true',
        help='Setup TestRail report client for upload test results.'
    )
    group.addoption(
        '--tr_deselect_tests',
        action='store_true',
        help='Deselect tests if its\' case is not present in run.'
    )
    parser.addini('tr_url', default='', help='Testrail Api url.')
    parser.addini('tr_user_email', default='', help='Testrail user\'s email.')
    parser.addini('tr_user_password', default='', help='Testrail user\'s password.')
    parser.addini('tr_project_id', default='', help='Testrail project id for new testrun creation.')


def pytest_configure(config):
    if config.getoption('--tr-reporting'):
        config.tr_config = TrConfig(config)
        config.tr_service = TrService(config)
        plugin = TrClient(config)
        config.pluginmanager.register(plugin)
        config.testrail_reporter = plugin



"""Config for setup package pytest plugin."""

import os

from setuptools import setup, find_packages

__version__ = '0.0.1b1'


def read_file(filename):
    """
    Read the given file.

    Args:
        filename: Filename to be read

    Returns:
        File content

    """
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return f.read()


setup(
    name='pytest-testrail-integrator',
    version=__version__,
    description='Pytest plugin for sending report to testrail system.',
    long_description_content_type="text/markdown",
    long_description=read_file('README.md'),
    author_email='vadym.stoilovskyi@gmail.com',
    url='https://github.com/VStoilovskyi/pytest-testrail-reporter',
    packages=find_packages(),
    install_requires=['pytest>=6.2.5', 'testrail-api'],
    license='Apache 2.0',
    keywords=['report', 'pytest', 'testrail'],
    classifiers=[
        'Framework :: Pytest',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    entry_points={
        'pytest11': [
            'pytest_testrail_integrator = pytest_testrail_integrator.plugin',
        ]
    }
)



import math
import pickle
from typing import List, Union

import pytest
from _pytest.config import ExitCode, Config
from _pytest.main import Session
from _pytest.mark import Mark
from _pytest.nodes import Item
from _pytest.reports import TestReport
from _pytest.runner import CallInfo

from .config import TrConfig
from .constants import TR_MARKER_NAME, TR_PASSED_TESTS_FLUSH_SIZE, TestrailStatus, \
    PytestStatus, TESTRAIL_STATUS_PRIORITY, PYTEST_TO_TESTRAIL_STATUS
from .dto import ReportDTO
from .service import TrService


def tr_case(case_id: Union[str, int]):
    return pytest.mark.case(case_id)


def is_master(config) -> bool:
    """
    True if the code running the given pytest.config object is
    running in a xdist master node or not running xdist at all.
    """
    return not hasattr(config, 'workerinput')


class TrClient:
    def __init__(self, config):
        self._service: TrService = config.tr_service
        self._config = config
        self._tr_config: TrConfig = config.tr_config
        self._results: List[ReportDTO] = []
        self._passed_tests_count = 0

    def pytest_collection_modifyitems(self, session: Session, config: Config, items: List[Item]):
        if self._service.is_test_run_available() and self._tr_config.deselect_tests:
            tr_run_cases = self._service.get_cases()
            deselected = []
            selected = []
            for item in items:
                marker = item.get_closest_marker(TR_MARKER_NAME)
                if not marker or int(marker.args[0]) not in tr_run_cases:
                    deselected.append(item)
                else:
                    selected.append(item)
            items[:] = selected

            config.hook.pytest_deselected(items=deselected)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item: Item, call: CallInfo):
        result: TestReport = (yield).get_result()
        if call.when == 'call' or result.outcome == 'skipped':
            if marker := item.get_closest_marker(TR_MARKER_NAME):
                case_id = int(marker.args[0])

                report = ReportDTO(
                    item.name,
                    item.nodeid,
                    item.own_markers,
                    PytestStatus(result.outcome),
                    result.duration,
                    result.longrepr,
                    case_id
                )

                self._results.append(report)

                if report.status == PytestStatus.PASSED:
                    self._passed_tests_count += 1

                # Try to flush passed results per node
                if self._service.is_test_run_available() and \
                        self._passed_tests_count > TR_PASSED_TESTS_FLUSH_SIZE:
                    self._try_send_passed_reports(self._results)
                    self._passed_tests_count = 0

    @pytest.hookimpl(hookwrapper=True, trylast=True)
    def pytest_sessionfinish(self, session: Session, exitstatus: Union[int, ExitCode]):
        yield
        if not is_master(session.config):
            # Serialize results and place in workeroutput to pass data to master node
            session.config.workeroutput['tr_results'] = pickle.dumps(self._results)
        else:
            if not self._service.is_test_run_available():
                # Todo: Probably it would be better to create new test run after all tests are collected.
                self._service.create_test_run(self._service.get_cases())

            self._service.upload_results(self._prepare_report(self._results, self._service.get_cases()))

    def pytest_testnodedown(self, node, error):
        """
        Extend master node results with worker's node results.
        """
        node_reports = pickle.loads(node.workeroutput['tr_results'])
        self._results.extend(node_reports)

    def _prepare_report(self, results: List[ReportDTO], testrun_cases: List[int]):
        # Remove redundant result reports
        filtered = filter(lambda x: x.case_id in testrun_cases, results)
        out = []
        for report_item in filtered:
            out.append(self.__prepare_item_report(report_item))

        # Sort results by testrail status.
        return sorted(out, key=lambda x: TESTRAIL_STATUS_PRIORITY[TestrailStatus(x['status_id'])])

    def __prepare_item_report(self, report: ReportDTO) -> dict:
        """Prepares API report item."""

        return {
            "case_id": report.case_id,
            "status_id": PYTEST_TO_TESTRAIL_STATUS[report.status].value,
            "comment": self.__prepare_comment(report),
            "elapsed": f"{math.ceil(report.duration)}s"
        }

    def __prepare_comment(self, report: ReportDTO):
        msg = report.name
        if self.__is_parametrized_test(report.markers):
            # crop by [ and ] to select parametrize ID
            msg = f"Test IDENTIFIER: {msg[msg.find('[') + 1: msg.find(']')]}"

        if report.status == PytestStatus.FAILED:
            return f'{msg}\n{report.longrepr.reprcrash.message}'
        if report.status == PytestStatus.SKIPPED:
            return msg + f'\n{report.longrepr[2]}'
        return msg

    @staticmethod
    def __is_parametrized_test(markers: List[Mark]):
        """
        Checks if the list contains parametrized marker.

        Args:
            markers: List of test's markers.

        Returns:
            boolean

        """

        return any([x.name == 'parametrize' for x in markers])

    def _try_send_passed_reports(self, results: List[ReportDTO]) -> None:
        """
        Prepare and send all passed tests only then delete them from the reports list.

        Args:
            results: All results within node.

        Returns:
            None

        """
        passed_results = list(filter(lambda x: x.status == PytestStatus.PASSED, results))
        results[:] = list(filter(lambda x: x.status != PytestStatus.PASSED, results))

        self._service.upload_results(self._prepare_report(passed_results, self._service.get_cases()))
